
/* A tiny BASIC compiler */

#include <stdint.h>
#include <stdio.h>
#include <stddef.h>
#include <setjmp.h>

#ifndef EOL
#define EOL	'\n'
#endif

int errors=0;

#define dbg(x...)	(DBG&&printf(x)||1)
int debug=0;
#define DBG	debug>0|(debug<0&&lineno==-debug)

#define isspace(c)	(c==' '||c=='\t')

enum { UNKNOWN, COMMAND, NUMBER, VARIABLE, DELIMITER, QUOTE, LINE };
char*en_tok[]={"UNKNOWN","COMMAND","NUMBER","VARIABLE","DELIMITER","QUOTE","LINE"};
 
enum { SPECIAL, OP, LD, VAR, NUM, ST, PC, LOC, GPUSH, GPOP , PRINT , SKIP, TEST, POKE, PEEK, BRANCH , LOOK, JUMP,
	FLYTO, FROMTO , COORD, CIRCLE , TURN , LOOKDOWN, LOOKUP, END, 
};
char*en_type[]={"SPECIAL","OP","LD","VAR","NUM","ST","PC","LOC","GPUSH","GPOP","PRINT","SKIP","TEST","POKE","PEEK","BRANCH","LOOK","JUMP","FLYTO","FROMTO","COORD","CIRCLE","TURN","LOOKDOWN","LOOKUP","END"};
	
enum { CIRCUIT, TAKEOFF, FLAIR };
char*en_spec[]={"CIRCUIT","TAKEOFF","FLAIR"};

char *prog;

static char*_mem_;
#define strdup(x)	(_mem_=strdup(x),_mem_||error(-1),_mem_)
char token[100];
int tok,token_type;
static char ahead;

int lineno=1;

putback()	{ ahead=1; }

static
struct opcode_t {
	int f;
	int x,y,z;
	char *s;
} opcode[900];
int pc=0;


emit0(int f,int x,int y, char* s) 
{opcode[pc].f=f;opcode[pc].x=x;opcode[pc].y=y;opcode[pc].s=s;dbg("EMIT: ");DBG&&popc(pc);return ++pc;}
#define emit(x,y)	emit0(x,y,0,0)

int inline my_tolower(char c) { return isupper(c)?tolower(c):c; }
static int _i;

#define lstack 100

#define lcmd 100
#define lvar 100
#define llbl 100
#define loop 100
#define lsym 100

char *lbls[llbl+1];
int   refs[llbl+1];
char *next[loop+1];
char *syms[lsym+1];
char *repl[lsym+1];
char *vars[lvar+1];
char *cmds[lcmd+1];
char *args[lcmd+1];
int (*func[lcmd+1])(int);

cmd(char *name, int (*f)(int), int arg) 
{
	if (insert(cmds,name,lcmd)<0) return;
	args[_i]=arg;
	func[_i]=f;
}

patch() 
{	int i;
	for(i=pc;i--;) 
		if (opcode[i].f==PC)
			opcode[i].f=LOC,opcode[i].x=refs[opcode[i].x];
	for(i=pc;i--;) 
		opcode[i].z=en_type[opcode[i].f];


}

popc(int i) {
	if (opcode[i].f<0)
	printf("0x%3.3x: %s %d %s\n",i,en_tok[-opcode[i].f],opcode[i].x,opcode[i].s?:""); else
	printf("0x%3.3x: %2.2d %-15.15s %s -> %d %d \t%c%s\n",i,opcode[i].f,en_type[opcode[i].f],opcode[i].f?"":en_spec[opcode[i].x],opcode[i].x,opcode[i].y,opcode[i].f==OP&&opcode[i].x?opcode[i].x:' ',opcode[i].s?:"");
}



dump() 
{	int i;
	printf("\n  Dump:\n--------\n");
	for(i=0;i<pc;i++) 
		popc(i);
}

int 
lookup(char *table[],char *item) 
{
	static void *old;
	if (table==old && table[_i] && !strcmp(table[_i],item)) return _i;
	_i=0; old=table;

	while(table[_i])
		if (*table[_i]==*item&&!strcmp(table[_i],item))
			return _i;
		else	_i++;
	return -1;
}

int 
insert(char *table[], char *item, size_t size) 
{
	lookup(table,item);
	if (_i<size) {
		item=strdup(item);
		if (!item) error(-1);
		table[_i]=item;
	} else	error("Table too small"),_i=-1;
	return _i;
}

jmp_buf jb;
char *argv0;

int main(int argc,char **argv)
{
  char *p_buf;
  int interactive=0;
  FILE *fp;
  char iline[80];

  for (argv0=*argv++,argc--; argc&&argv[0][0]=='-'; argc--,argv++) 
	  if (argv[0][1]=='#') { debug++; if (argv[0][2]==':') debug=-strtol(&argv[0][3],NULL,0);}  else 
	  if (argv[0][1]=='@') debug++; else
		  ;

  if (argc!=1) fprintf(stderr,"Error: invalid arguments\n Usage: %s <file>\n",argv0),exit(-1);

  p_buf=malloc(16384);
  memset(p_buf,0xff,16384);
  p_buf[0]=0;
  
  fp=fopen(argv[0],"r");
  if (!fp) fprintf(stderr,"Unable to open input file %s\n",*argv),exit(1);
  { int i=0; char *p; p=p_buf;
  while((*p=fgetc(fp)),i++<16384-2&&!feof(fp)) *p&=127,p++;
  *p++='\n'; *p++='\0';
  fclose(fp);
  }

// restart
  
  prog = p_buf;
  init_cmds();
  find_labels();

  setjmp(jb);
  while(dbg("----------------------------\n")&&get_token()) { char *p;
    //p=prog; printf("LINE(%d): %s %s\n",lineno,token,get_eol()); prog=p;
    if(token_type==VARIABLE&&*prog==':'||token_type==NUMBER) {
	    if (lookup(lbls,token)<0) error("label error"); else refs[_i]=pc;
	    if (token_type==VARIABLE) prog++; continue;
    } else
    if(token_type==VARIABLE) {
      putback(); 
      assignment(); 
    } else
    if (token_type==COMMAND) {
	    putback();
	    command();
    } else
	    find_eol();
  }

  emit(END,0);
  patch();
  dump();
  free(p_buf);
  return 0;
}

command() {
	get_token();
    if (token_type==COMMAND) {
	    emit0(-COMMAND,args[tok],0,strdup(token));
	    if (func[tok])
		    func[tok](args[tok]);
	    else	emit(args[tok],0);
	    find_eol();
	    return 1;
    } else	putback();
    return 0;
}


find_labels() {
	char *temp; int i; i=debug; debug=0;
	temp=prog;
	while(get_token()) {
		if (token_type==NUMBER||token_type==VARIABLE&&*prog==':') {
			if (lookup(lbls,token)<0)
				insert(lbls,token,llbl);
			else	fprintf(stderr,"Error(%d): double defined label `%s'\n",lineno,token),errors++;
		}
		find_eol();
	}
	prog=temp; debug=i;
	lineno=1;
}

find_eol() { if (token_type==DELIMITER&&tok==EOL) return ahead=0; while(get_token()&&*token!='\n'); }
#define lbuff	1024
char *
get_eol() {
	static char buff[lbuff+2]; int i; char *p; p=buff;
	while(*prog&&*prog!='\n')  *p++=*prog++;
	*p++='\n'; *p='\0';
	dbg("TOKEN: %s",buff);
	return buff;
}



/* Get a token. */
int get_token()
{
  static int stack[lstack+1];
  register char *temp;
  static int old,flag;

  if (ahead) { ahead=0; dbg("!"); goto ret0; };

  if (flag) lineno++,flag=0;

  if (old!=lineno) {  char *p;
	  old=lineno; temp=prog;
	  dbg("Line(%d): ",lineno); 
	  emit0(-LINE,lineno,0,strdup(get_eol()));
	  prog=temp;
  }

  while(isspace(*prog)) prog++;  /* skip over white space */

  token_type=0; tok=0;
  temp=token;

  if(*prog=='\'') while(*prog&&*prog!='\n') prog++;	// skip comment

  if(!*prog) { /* end of file */
    tok = EOF; 
    if (stack[0])  prog=stack[stack[0]--]; else goto ret;
    return get_token();
  }

  if(isnumber(*prog)) { /* number */
    while(!isdelim(*prog)) *temp++=my_tolower(*prog++);
    *temp='\0';
    token_type=NUMBER;
    if (token[0]=='0') {
    if (token[1]|' '=='x'&&(tok=16)||token[1]|' '=='b'&&(tok=2))	tok=strtol(token+2,NULL,tok);
    } else 
    if (token[0]=='$'&&(tok=16)||token[0]=='%'&&(tok=2))		tok=strtol(token+1,NULL,tok);
      else								tok=strtol(token,NULL,10);
    if (*token=='%'&&token[1]=='\0') token_type=DELIMITER;
    goto ret;
  }

  if(*prog=='"') { /* quoted string */
    token_type=QUOTE;
    while(*prog&&*++prog!='"'&&*prog&&*prog!='\n') *temp++=*prog;
    if(*prog=='\n') error("unterminated string");
    *prog++;
    goto ret;
  }
  
  if(isalpha(*prog)) { /* var or command */
    while(!isdelim(*prog)) *temp++=my_tolower(*prog++);
    *temp = '\0';
    tok=lookup(cmds,token);
    if(tok<0) { token_type = VARIABLE;
	    tok=lookup(vars,token);
	    if (tok<0&&lookup(syms,token)!=-1)  {
		    dbg("Sym %s\n",token);
		    if (stack[0]==lstack) error("out of stack"); else stack[0]++;
		    stack[stack[0]]=prog;
		    prog=repl[_i];
		    return get_token();
	    }
    }
    else token_type = COMMAND; /* is a command */
    goto ret;
  }

    token_type=DELIMITER;
    if ((tok=*temp++=*prog++)=='\n') flag++;
    if (tok==1) *token=tok=EOL;

  while(isspace(*prog)) prog++;  /* skip over white space */


ret:
  *temp = '\0';
  while(isspace(*prog)) prog++;  /* skip over white space */
ret0:
  dbg("Token(%d): %-15s %-20s %d \n",lineno,en_tok[token_type],*token==EOL?"<EOL>":token,tok);
  return token_type;
}



int isdelim(char c)  {return !(isalpha(c)||isnumber(c)||c=='_'); }
int isnumber(char c) { return (isdigit(c))||(c=='$' || c=='%');}

void get_exp() {dbg("enter %s\n",__FUNCTION__);get_token();if (!*token) error("missing expression");level2();putback();dbg("exit %s\n",__FUNCTION__);}

#define exp(l,n,m,x,y)	\
void level##l(){register char o;dbg("enter %s\n",__FUNCTION__);level##n();while((o=*token)==x||o==y){get_token();level##m();emit(OP,o);}}
#define exp_(l,x,y,a,e,v,t)	\
void level##l(){register char o=0;dbg("enter %s\n",__FUNCTION__);if(*token==x||*token==y){o=*token;get_token();a;}else e;v;if(o)t;}

exp(2,3,3,'+','-')
exp(3,4,4,'*','/'||o=='%')
exp_(4,'m'&&!strcmp(token,"max"),'m'&&!strcmp(token,"min"),o=token[1]=='a'?'M':'m',,level5(),emit(OP,o))
exp_(5,'+','-',,,level6(),unary(o))
exp_(6,'(','(',level2();if(*token!=')')error("parentesis expected");get_token(),primitive(),,)


void primitive(){dbg("enter %s\n",__FUNCTION__);if(token_type==VARIABLE&&tok==-1)error("Variable don't exist");
	if (token_type==VARIABLE&&emit(VAR,tok)||token_type==NUMBER&&emit(NUM,tok));else error(0);get_token();}
void unary(char o){dbg("enter %s\n",__FUNCTION__);if(o=='-')emit(OP,'u');}



error(l) {
	if (!l) 	return error("Syntax error");
	if (l==-1) 	return error("Out of memory");
	ahead=0;
	fprintf(DBG?stdout:stderr,"Error(%d): %s\n",lineno,l);
	if (++errors==10) error("too many errors, unable to continue"),exit(1);
	find_eol();
	longjmp(jb,1);
}


/* end of the core BASIC compiler */

req(char c)	{ if (get_token()!=DELIMITER||*token!=c) error(0); }

variable() {
	if (get_token()!=VARIABLE) error("variable required");
	if (tok<0)	tok=insert(vars,token,lvar);
	return tok;
}

assignment() {
	int var;
	dbg("enter %s\n",__FUNCTION__);
	var=variable();
	req('=');
	get_exp();
	emit(ST,var);
}

gosub(l) {
	if (l) emit(GPUSH,0);
	if (get_token()!=VARIABLE&&token_type!=NUMBER) error("required a valid label");
	if (lookup(lbls,token)<0) error("not a valid label");
	emit(PC,_i);
}

done() {
	if (token_type!=DELIMITER&&tok!=EOL) find_eol();
}

do_for(l) {
	int var,loc; 
	char tmp[100];
	char buff[1024]; char *p=&buff[0];
	static int last=-1;
	dbg("enter %s\n",__FUNCTION__);
	var=variable();
	if (!l) {
	strcpy(tmp,token); putback();
	assignment();
	//get_token();
	sprintf(buff,"%d_for_loop",var); tok=lookup(lbls,buff); 
	if (tok>=0) for(loc=0;loc<pc;loc++) if (opcode[loc].f==PC&&opcode[loc].x==tok) opcode[loc].f=LOC,opcode[loc].x=refs[opcode[loc].x];
	if (tok<0) tok=insert(lbls,buff,llbl);
	loc=tok;
	if (token_type!=VARIABLE||strcmp(token,"to")) error("syntax error in for statement");
                            //FOR  variable = start TO end {STEP {-}increment}
			    //++++hack
	sprintf(buff,"%s = %s + 1\1 if %s <> %s then goto %d_for_loop",tmp,tmp,tmp,get_eol(),var);
	p=strdup(buff); if (!p) error(-1);next[var]=p;
			    //----hack
	return;
	p+=sprintf(buff,"if %s <> ",tmp);
	while(	get_token() ) {
		if (token=='\n') break;
		if (!strcmp("step",token)) break;
		p+=sprintf(p," %s", token);
	}
	*p++='\n'; *p='\0';
	p=get_eol();
	strcat(p,buff);
	sprintf(buff,"%s = %s + %s then goto %d_for_loop \n",tmp,tmp,p,var);
	p=strdup(buff); if (!p) error(-1);
	next[var]=p;
	} else {
	if (var<0) error("variable is not valid");
	if (!next[var]) error("not a valid for loop");
	p=prog; prog=next[var];
	assignment(); command();
	prog=p;
	}
}

todo(char *str)	{ printf("TODO(%d): %s \n",lineno,str,lineno); emit(SKIP,0); }

cmd_todo(char *str)	{	todo(str); find_eol(); emit(SKIP,0); }

cond() {
	int val1,val2; int c;
	dbg("enter %s\n",__FUNCTION__);
	get_exp();
	if (get_token()==VARIABLE&&!strcmp("then",token)) {
		emit(NUM,0);
		emit(OP,'!');
		putback();
		return;
	}
	if (token_type!=DELIMITER) error(0);
	c=*token; if (!strchr("<>=",c)) error(0);
	if (get_token()==DELIMITER) 
	switch(*token) {
		case '>': if (c=='<') c='!'; else error(0); break;
		case '=': if (c=='<') c='{'; else if (c=='>') c='}'; else error(0); break;
		default: putback();
	}	else 	 putback();
	get_exp();
	emit(OP,c);
	if (token_type==VARIABLE&&(!strcmp(token,"and")||!strcmp(token,"or"))) {
		get_token();
		c=*token;
		cond();
		emit(OP,c);
	}
}
do_if(l) {
	int val1,val2; int c;
	dbg("enter %s\n",__FUNCTION__);
	cond();
	emit(TEST,0);
	get_token();
	if (token_type!=VARIABLE||strcmp("then",token)) error(0);
	longjmp(jb,3);
}

dosym(char *sym) {
	if (get_token()!=VARIABLE||tok!=-1) error(0);
	if (lookup(syms,token)!=-1) error("symbol already defined");
	if (*prog!='=') error(0); else prog++;
	if (insert(syms,token,lsym)!=-1)
		repl[_i]=strdup(get_eol());
		
}

coord() {
	int i=0;
	dbg("enter %s\n",__FUNCTION__);
	req('(');
	get_exp(),i++;
	if (*token=='.') get_exp(),i++;
	req(',');
	get_exp(),i++;
	if (*token=='.') get_exp(),i++;
	req(')');
	if (i&1) error(0);
	emit(COORD,i);
}

/*   bset(emit1,emit2,emit3,count1,count2,flag1_only_vars,flag2_emit_EOF) */
#define bset(a,b,c,c1,c2,m1,m2)	((((a)&0xff)<<24)|(((b)&0xff)<<8)|(((c)&0xff)<<0)|(((c1)&0x7)<<21)|(((c2)&0x7)<<18)|(((m1)&0x1)<<16)|(((m2)&0x1)<<17))
bargs(l) {
	int i;
	dbg("enter %s\n",__FUNCTION__);
	i=(l>>24)&0xff;
	if (i) emit(i,0);
	i=(l>>16)&0xff;
	i>>=5;i&=7;
	while(i--) {
		get_exp();
		req(',');
	}
	i=(l>>8)&0xff;
	if (i) emit(i,0);
	req('(');
	do get_exp(); while (*token!=')');
	req(')');
	i=(l>>16)&0xff;
	i>>=2;i&=7;
	if ((l>>16)&2) emit(-1,0);
	while(i--) {
		req(',');
		if 	((l>>16)&1)
			variable();
		else 
			get_exp();
	}
	i=(l&0xff);
	emit(i,0);
}


cargs(l) {
	int i; i=l&0xff;
	dbg("enter %s\n",__FUNCTION__);
	while(i--) {
		coord();
		if (i) req(',');
	}
	emit(l>>8,i);
}

xargss(l) {
	int i; i=l&0xff;
	dbg("enter %s\n",__FUNCTION__);
	while(i--) {
		get_exp();
		if (i) req(',');
	}
	emit(l>>8,i);
}

#define xset(a,b,c,m)	((((a)&0xff)<<8)|(((b)&0xff)<<0)|(((c)&0xff)<<24)|(((m)&0xff)<<16))
xargs(unsigned int l) {
	int i,t;
	dbg("enter " __FUNCTION__ "\n");
	i=(l>>24)&0xff;
	t=(l>>16)&0xff;
	while(i--) {
		if (t&(1<<i))
			variable();
		else	get_exp();
		if (i)	req(',');
	}
	i=(l>>8)&0xff;
	t=(l>>0)&0xff;
	dbg("emitting: %d - %d\n",i,t);
	dbg("exit " __FUNCTION__ "\n");
	emit(i,t);
}

do_end(l) {
	dbg("enter %s\n",__FUNCTION__);
	get_token();
	if (!strcmp(token,"sub"))	emit(GPOP,0);	else
	if (!strcmp(token,"pattern"))	emit(GPOP,0);	else
	if (!strcmp(token,"thread"))	emit(GPOP,0);	else
	putback(), emit(END,0);
	error(0);
}

do_define(l) {
	dbg("enter %s\n",__FUNCTION__);
	if (get_token()!=VARIABLE||tok!=-1) error(0);
	if (lookup(lbls,token)>=0) error("symbol already defined");
	insert(lbls,token,llbl); refs[_i]=pc;
	get_token();
	if (strcmp(token,"is")) error(0);
}

init_cmds() 
{
	cmd("symbol",dosym,token);
	cmd("for",do_for,0);
	cmd("next",do_for,1);
	cmd("if",do_if,0);
	cmd("goto",gosub,0);
	cmd("gosub",gosub,1);
	cmd("return",(void*)0,GPOP);
	cmd("peek",xargs,xset(PEEK,0,2,1));
	cmd("poke",xargs,xset(POKE,0,2,0));
/*   bset(emit1,emit2,emit3,count1,count2,flag1_only_vars,flag2_emit_EOF) */
	cmd("branch",bargs,bset(BRANCH,0,JUMP,1,0,0,0));
	cmd("lookdown",bargs,bset(LOOKDOWN,0,0,1,1,1,0));
	cmd("lookup",bargs,bset(LOOKUP,0,0,1,1,1,0));
	cmd("flyto",cargs, FLYTO << 8 | 1);
	cmd("fromto",cargs, FROMTO << 8 | 2);
	cmd("circle",xargs, xset(CIRCLE,0,2,0));
	cmd("turn",xargs, xset(TURN,0,2,0));
	cmd("takeoff",xargs,xset(SPECIAL,TAKEOFF,0,0));
	cmd("circuit",xargs,xset(SPECIAL,CIRCUIT,0,0));
	cmd("flair",xargs,xset(SPECIAL,FLAIR,0,0));
	cmd("define",cmd_todo,0);
	cmd("end",do_end,0);
	cmd("sub",do_define,0);
}
