#include <stdint.h>
#define OK	0xff
#define FAIL	0xf7

#define EE_OFFSET 100

#ifdef USE_PINZ
pinz(uint8_t p) { return (1<<(p&0xf));}
#else
#define pinz(x)	(1<<(x))
#endif

#define b(x)	+(1<<(x))
#define IMM	b(0)b(1)b(2)b(3)b(4)

#define OFFSET  16
#define START	800
#define BEGIN	4

#define STACK	20
#define HWSTACK	4

enum { INT, MASK, WAIT, SYS };

#define ERR_UNDERFLOW ~0x80
#define ERR_OVERFLOW  ~0x40
#define ERR_DIV_ZERO  ~0x20

#define ERR_CRITICAL	(OK&ERR_UNDERFLOW&ERR_OVERFLOW)

typedef
struct self_s {
	uint8_t  error;
	uint16_t vars[0xf];
	uint16_t pc;
	uint16_t tick;
	uint16_t stack[STACK];
	uint16_t sys[SYS];
	uint16_t hwstack[HWSTACK];
	int8_t	 sp,hwsp;
} *self_t;

struct self_s vm[1];

#define pc	self->pc
#define sp	self->sp
#define hwsp	self->hwsp
#define stack	self->stack
#define hwstack	self->hwstack
#define vars	self->vars
#define sys	self->sys
#define error	self->error
#define tick	self->tick

#define  fetch(x)	((eeprom_read((x)+BEGIN)<<8)|eeprom_read((x)+BEGIN+1))

#define push(x)		({ if (sp==STACK) return ERR_OVERFLOW; stack[sp++]=x; })
#define pop()		(SP(1),stack[sp])
#define hwpush(x)	({ if (hwsp==HWSTACK) return ERR_OVERFLOW; hwstack[hwsp++]=x; })
#define hwpop()		({ if (!hwsp) return ERR_UNDERFLOW; hwstack[--hwsp]; })
#define hwTOS(x)	hwstack[hwsp-1]
#define TOS(x)		stack[sp-1]	// TICKS	if no data is on stack
#define NOS(x)		stack[sp-2]	// PC		if no data is on stack
#define ST(x)		({if (sp -x<0) return ERR_UNDERFLOW; })
#define SP(x)		(ST(x),sp-=x)

static
uint16_t global[0xf]; 

static uint16_t tos,nos;
static uint8_t  arg,flag;

#define fastcall	static	// todo

#define OPC(add,x...)	fastcall uint8_t opc_##add(self_t self) { x; return OK; }
#define min(x,y)	(x<y?x:y)
#define max(x,y)	(x>y?x:y)

OPC(load,tos=eeprom_read(pc++)<<8 | eeprom_read(pc++); push(tos))

OPC(dummy,return OK)	

#define math(name,o,a...)	OPC(name,SP(2),nos = nos o tos,push(nos),##a)

math(add,+) math(sub,-) 
math(mul,*,0;if(!nos||!tos) return ERR_DIV_ZERO;) 	math(div,/,0;if(!nos||!tos) return ERR_DIV_ZERO;) 	
math(mod,%,0;if(!nos||!tos) return ERR_DIV_ZERO;) 	
math(lt,<) math(gt,>) math(le,<=) math(ge,>=) 
OPC(neg,ST(1),TOS()=!tos)
OPC(eq,opc_sub(self),opc_neg(self))

OPC(min,SP(2),nos=max(nos,tos),push(nos))
OPC(max,SP(2),nos=min(nos,tos),push(nos))

OPC(peek,TOS()=getval(tos))	OPC(poke,setval(nos,tos),SP(2))	OPC(read,TOS()=eeprom_read2((tos&0x1f)+EE_OFFSET)) 
	OPC(write,eeprom_write2((nos&0x1f)+EE_OFFSET,tos),SP(2))
OPC(goto,pc=pop())		OPC(return,pc=hwpop())
	OPC(mark,hwpush(pc))	OPC(drop,hwpop())		OPC(lgoto,opc_load(self),opc_goto(self))
OPC(load2,opc_load(self),opc_load(self)) OPC(mask,opc_load2(self),tos=pop(),nos=pop(),sys[0]=tos,sys[1]=nos)

OPC(do_branch,arg=tos;SP(2);flag=0;if (nos>tos) { pc+=arg+arg; return OK; } pc+=nos+nos; opc_load(self); arg-=nos; pc+=arg+arg;flag=1)
OPC(lookdown,arg=tos;SP(2);tos=0;flag=0;while(arg--) {opc_load(self); if (stack[sp--]==nos) tos=flag; flag++;} push(tos))
OPC(branch,opc_do_branch(self); if(flag) opc_goto(self))
OPC(lookup,opc_do_branch(self); if(!flag) push(0))

OPC(load4,opc_load2(self),opc_load2(self))

OPC(dup,push(tos)) OPC(swap,SP(2),push(tos),push(nos)) OPC(pop,pop())
	OPC(redo, nos=hwpop(); if(tos) { hwpush(nos); push(nos);opc_goto(self);})

OPC(wait,sys[WAIT]=pop())

#undef OPC
#define OPC(a,b,c,d)	opc_##a, opc_##b, opc_##c, opc_##d

uint8_t (*opcode[])(void*) = {
	opc_dummy, opc_dummy,		// have full access to 4bit arg value + 1bit flag value or 5bit* arg value, 
					// depending on IMM. it's  the bit 3 and 4
	OPC( add, sub, mul, div),	// and is a alias to mul, or is a alias to add	
	OPC( lt, gt, ge, le),
	OPC( eq, neg, min, max),	// ne is a alias to sub
	OPC( peek, poke, read, write ),

	OPC( redo, return, mark, drop),	// works on hw stack
	OPC( dup, swap, pop, load),	// works on sw stack
	OPC( load2, load4, lgoto, goto),
	OPC( branch, lookup, lookdown, mask ),
	/*
	OPC( in , ins, out, outs ),
	*/
	/* DSL begins here */
	OPC( wait, dummy, dummy, dummy )
};

uint8_t
interpreter(self_t self) {
	register uint8_t opc; arg=0;
	if (!pc) pc=8;
	tick++;
	if (sys[WAIT]) { sys[WAIT]--; return OK; }
	opc=eeprom_read(pc++);
	arg=opc&0x1f; flag=opc&0x10; opc>>=5;
	if (!(pinz(opc)&IMM))	arg&=0xf;
	if (!opc)	{ pc-=arg; return OK; }
	if (opc<5)	{ flag=opc; if (opc<3) { while(--flag) arg+=32; arg+=2; } 
			  tos=TOS(); nos=NOS(); return error&=opcode[opc-3](self); 
			}
	opc-=5;
	if (flag)	{
	if (!opc--)	{ opc_mark(self); pc=fetch(arg); return OK; }
	if (!opc--)	{ if (!pop()) pc+=arg; return OK; }
	if (!opc--)	{ nos=vars[arg]; push(nos); return OK; }
	} else {
	if (!opc--)	{ nos=arg; if (arg>=8) { nos=(arg&7)<<8; nos|=eeprom_read(pc++); } return push((arg)),OK; }
	if (!opc--)	{if(arg&8){nos=NOS(--);global[arg&7]=nos;return OK;}nos=global[arg&7];push(nos);return OK;} 
	if (!opc--)	{ nos=pop(); vars[arg]=nos; return OK; }
	}
}

main() { interpreter(&vm[0]); }
eeprom_read(uint16_t a) { return ~a;}
eeprom_write(uint16_t a,uint16_t b) { return ~a;}
eeprom_read2(uint16_t a) { return eeprom_read(a)<<8|eeprom_read(a+1); }
eeprom_write2(uint16_t a,uint16_t b) { return eeprom_write(a,b>>8)<<8|eeprom_write(a+1,b&0xff); }
getval(int a) { return ~a;}
setval(int a,int b) { return ~a;}
